#include <iostream>
#include <cstdlib>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <memory.h>
#include <unistd.h>
#include <time.h>

#include "discover_lib.h"

using namespace std;

/**********************************************************************
 * summary : Constructer (prepare socket)
 * return  : -
 *********************************************************************/
CDiscover::CDiscover()
{
	m_sock = -1;
}
/**********************************************************************
 * summary : Setup socket 
 * return  : -
 *********************************************************************/
bool CDiscover::Setup()
{
	////
	// Set up the address we're going to bind to.
	////
	if(m_sock < 0){
		close(m_sock);
	}

	struct sockaddr_in name;
	name.sin_family = AF_INET;
	name.sin_port = htons(CLIENT_PORT);
	name.sin_addr.s_addr = INADDR_ANY;
	memset (name.sin_zero, 0, sizeof (name.sin_zero));

	// Make a socket...
	if((m_sock = ::socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
		cerr << "Can't create dhcp socket" << endl;
		return false;
	}

	// Set the REUSEADDR option so that we don't fail to start
	int flag = 1;
	if(::setsockopt(m_sock, SOL_SOCKET, SO_REUSEADDR,
			(char *)&flag, sizeof flag) < 0){
		cerr << "Can't set SO_REUSEADDR option on dhcp socket" << endl;
		return false;
	}

	// Set the BROADCAST option so that we can broadcast DHCP responses.
	if(::setsockopt(m_sock, SOL_SOCKET, SO_BROADCAST,
            (char *)&flag, sizeof flag) < 0){
		cerr << "Can't set SO_BROADCAST option on dhcp socket" << endl;
		return false;
	}

	// Bind the socket to this interface's IP address.
	if(::bind(m_sock, (struct sockaddr *)&name, sizeof name) < 0){
		cerr << "Can't bind to dhcp address" << endl;
		return false;
    }

	////
	// prepare broadcast address
	////
	memset((char *)&m_sockaddr_broadcast, 0, sizeof(struct sockaddr_in));
	m_sockaddr_broadcast.sin_family = AF_INET;
	m_sockaddr_broadcast.sin_port = htons(SERVER_PORT);
	m_sockaddr_broadcast.sin_addr.s_addr = INADDR_BROADCAST;

	return true;
}

/**********************************************************************
 * summary : send indicated packet
 * return  : -
 *********************************************************************/
CDiscover::~CDiscover()
{
	if(m_sock >= 0){
		close(m_sock);
	}
}

/**********************************************************************
 * summary : Make new transaction ID
 * return  : transaction ID
 *********************************************************************/
unsigned long CDiscover::GetNewTransactionID() const
{
	while(1){
		srand((unsigned)time(NULL));
		unsigned long num = (unsigned long)rand();
		if(num){
			return num;
		}
	}
}

/**********************************************************************
 * summary : send indicated packet
 * return  : -
 *********************************************************************/
bool CDiscover::SendPacket(const discover_packet_s* packet)
{
	int send = ::sendto(
						m_sock,
						(char*)packet,
						sizeof(discover_packet_s),
						0,
						(struct sockaddr*)&m_sockaddr_broadcast,
						sizeof(m_sockaddr_broadcast)
					);
	return (send == sizeof(discover_packet_s));
}

/**********************************************************************
 * summary : Broadcast find packet
 * return  : transaction id
 *********************************************************************/
unsigned long CDiscover::SendFindPacket()
{
	discover_packet_s packet;
    memset(&packet, 0, sizeof(discover_packet_s));
    packet.message_type = (message_type_e)htonl(DISCOVER_REQ);
    packet.transaction_id = GetNewTransactionID();
	if(SendPacket(&packet)){
		return packet.transaction_id;
	}
	
	return 0;
}

/**********************************************************************
 * summary : Broadcast config request packet
 * return  : transaction id
 *********************************************************************/
unsigned long CDiscover::SendConfigPacket(
	const char* macaddr,
	const network_info_s* networkInfo
)
{
	discover_packet_s packet;
    memset(&packet, 0, sizeof(discover_packet_s));
    packet.message_type = (message_type_e)htonl(IP_OFFER_REQ);
    packet.transaction_id = GetNewTransactionID();
	memcpy(&packet.mac_addr, macaddr, 6);
	memcpy(&packet.net_info, networkInfo, sizeof(network_info_s));
	if(SendPacket(&packet)){
		return packet.transaction_id;
	}
	
	return 0;
}

/**********************************************************************
 * summary : Receive Packet
 * return  : -
 *********************************************************************/
bool CDiscover::RecvPacket(
	discover_packet_s* packet
)
{
	struct sockaddr sender;
	socklen_t len = sizeof(sender);
	int size = ::recvfrom(
						m_sock,
						(char*)packet,
						sizeof(discover_packet_s),
						0,
						&sender,
						&len
					);
	return size == sizeof(discover_packet_s);
}

/**********************************************************************
 * summary : Receive packet until timeout
 * return  : -
 *********************************************************************/
bool CDiscover::RecvReplyPacket(
	discover_packet_s* packet,
	unsigned long transactionID,
	unsigned int timeout
)
{
	struct timeval* tout = NULL;
	if(timeout){
		tout = new struct timeval;
		tout->tv_sec = timeout;
		tout->tv_usec = 0;
	}
		
	while(1){
		fd_set readfds;
		FD_ZERO(&readfds);
		FD_SET(m_sock, &readfds);

		switch(select(FD_SETSIZE, &readfds, NULL, NULL, tout)){
		case 0:
			// timeout
		case -1:
			// error;
			return false;
		default:
			break;
		}

		if(!FD_ISSET(m_sock, &readfds)){
			cerr << "FD is not set" << endl;
			return false;
		}

		if(!RecvPacket(packet)){
			return false;
		}

		if(packet->transaction_id == transactionID){
			return true;
		}
	}
}
